/*
 * IBM Performance Inspector
 * Copyright (c) International Business Machines Corp., 1996 - 2008
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef _BPUTIL_COMMON_H_
#define _BPUTIL_COMMON_H_

   #define TRACE_RECORD_HEADER  12  // 12 bytes header for each record: TY-LEN-MAJ:MIN:TS

#ifndef STAP_ITRACE // DEFS
   #define MTE_TYPE              1
   #define TRACE_TYPE            2
   #define MTE_FINAL_TYPE        4
#endif // STAP_ITRACE DEFS

   #define BP_OK     0
   #define BP_ERROR -1

#define TRACE_BUFFER_FULL   0x00000001   /* Trace buffer full condition */


/**********************************************************************************/
#ifndef STAP_ITRACE // DEFS
typedef struct _PerfHeader
{
   UCHAR   nrm2[4];                         //   0: "NRM2"
   UINT32  endian;                          //   4: 0x040302010
   UCHAR   cr;                              //   8: Carriage return
   UCHAR   pad[3];                          //   9: Padding
   UINT32  header_size;                     //  12: Size of PerfHeader - set to 512
   UINT32  data_size;                       //  16: Size of data in buffer (that follows header)
   UINT32  buffer_type;                     //  20: ***** Not used and set to zero *****
   UINT32  cpu_no;                          //  24: CPU number to which this buffer belongs
   UINT32  no_ts_hooks;                     //  28: 1=Hooks of type 01 have no timestamp
   UINT32  collection_mode;                 //  32: NORMAL, WRAP, CONTINUOS, MTE_END
   UINT32  low_start_time;                  //  36: Time of last Trace on
   UINT32  hi_start_time;                   //  40:
   UINT32  low_end_time;                    //  44: Time of last Trace off
   UINT32  hi_end_time;                     //  48:
   UINT32  Metric_cnt;                      //  52: Number of metrics
   UINT32  flag;                            //  56: 1=At least one trace buffer full
   UINT32  num_cpus;                        //  60: Number of CPUs
   UINT32  uncompressed_data_size;          //  64: Size of uncompressed data
   UINT32  OS_type;                         //  68: Whether 64 or 32 bit OS
   UINT32  inston;                          //  72: Instruction count on/off in itrace
   UINT32  interval_id;                     //  76: interval id at last trace on or zero
   UINT64  kallsyms_offset;                 //  80: (Linux-only) Offset to kallsyms file data
   UINT64  kallsyms_size;                   //  88: (Linux-only) Size of kallsyms file data
   UINT64  modules_offset;                  //  96: (Linux-only) Offset to modules file data
   UINT64  modules_size;                    // 104: (Linux-only) Size of modules file data   
   UINT32  datalosscount;                   // 108: (Linux-only) lost hooks when continous tracing 
   UINT32  section_type;                    // 112: (Linux-only) TRACE, MTE, or FINAL_MTE type section
   
   // used by the driver only
   
   UINT32  buffer_size;                     // 116
   UINT32  ulUpperCycles;                   // 120

   // offsets into the buffer
   
   UINT32  buffer_start;
   UINT32  buffer_end;
   UINT32  buffer_current;
   UINT32  buffer_oldest;
   
   // for continuous tracing
      
   UINT32  buffer_rd_offset;                // Offset to be read next 
   UINT32  buffer_varend;                   // When the last record does not fit
   UINT32  wrap_rd;                         // rd offset wrapped
   UINT32  rd_behind_wr;                    // rd behind wr (rd < wr)
    
}PERFHEADER;

typedef PERFHEADER *PPERFHEADER;
#define PERFHEADER_BUFFER_SIZE 512

#define DEFINED_CRETURN 10
#define LINUX64_FLAG 0x64
#define LINUX32_FLAG 0x32


/************************ Trace Hook Control Blocks ************************/

#define NUM_RAWINT32 125

typedef struct RAWINT_HOOK_32BIT {
        uint16_t  typeLength;
        uint16_t  majorCode;
        uint32_t  minorCode;
        uint32_t  timeStamp;
        uint32_t  intval[NUM_RAWINT32];
}RAW32_HOOK;

#define NUM_INT32_SMALL_HOOK 5

typedef struct RAWINT_SMALL_HOOK_32BIT {
        uint16_t  typeLength;
        uint16_t  majorCode;
        uint32_t  minorCode;
        uint32_t  timeStamp;
        uint32_t  intval[NUM_INT32_SMALL_HOOK];
}RAW32_SMALL_HOOK;

#define SMALL_HOOK_SIZE sizeof(RAW32_SMALL_HOOK)

#define HOOK_HEADER_SIZE (2 * sizeof(uint16_t)) + (2 * sizeof(uint32_t))
#endif // STAP_ITRACE DEFS

typedef struct _STRUCT_DISPATCH_HOOK
{
        uint16_t  typeLength;
        uint16_t  majorCode;   /* Major code 0x12 */
        uint32_t  minorCode ;  /* Minor code      */
        uint32_t  timeStamp;
        uint32_t  parent ; 
}DISPATCH_HOOK;
#define DISPATCH_HOOK_SIZE sizeof(DISPATCH_HOOK)

typedef struct STRUCT_IRQ_HOOK_32BIT {
        uint16_t  typeLength;
        uint16_t  majorCode;      /* Major code 0x04 */
        uint32_t  irq;            /* ORed with 0x80000000 for exit */
        uint32_t  timeStamp;
	uint32_t  count;
}IRQ_HOOK;
#define IRQ_HOOK_SIZE sizeof(IRQ_HOOK)

typedef struct STRUCT_TS_CHANGE_HOOK {
    uint16_t  typeLength;
    uint16_t  majorCode;
    uint32_t  minorCode;      
    uint32_t  timeStamp;
    uint32_t  timeStamp_hi;
}TS_CHANGE_HOOK;
#define TS_CHANGE_HOOK_SIZE sizeof(TS_CHANGE_HOOK)

typedef struct STRUCT_TPROF_HOOK_32BIT {
        uint16_t typeLength;
        uint16_t majorCode;     /* Major code 0x10 */
        uint32_t  minorCode;    /* Minor 10 + cpl */
        uint32_t  timeStamp;
        uint32_t  pid;
        uint32_t  addrl;
#ifdef CONFIG_64BIT
	uint32_t  addrh;
#endif
}TPROF_HOOK;
#define TPROF_HOOK_SIZE sizeof(TPROF_HOOK)

#ifndef STAP_ITRACE // DEFS
typedef struct STRUCT_MTE_HOOK_32BIT {
        uint16_t typeLength;
        uint16_t majorCode;
        uint32_t  minorCode;
        uint32_t  timeStamp;
        uint32_t  int_cnt;
        uint32_t  pid;
        uint32_t  code_size;
        uint32_t  NT_ts;
        uint32_t  NT_cs;
        uint32_t  addrl;
#ifdef CONFIG_64BIT
	uint32_t  addrh;
#endif
        uint32_t  str_len;
        char      str_val[256];
}MTE_HOOK;
#define MTE_HOOK_SIZE (sizeof(MTE_HOOK) - 256)

typedef struct STRUCT_PID_NAME_HOOK_32BIT {
        uint16_t  typeLength;
        uint16_t  majorCode;
        uint32_t  minorCode;
        uint32_t  timeStamp;
        uint32_t  int_cnt;
        uint32_t  pid;
        uint32_t  str_len;
        char      str_val[256];
}PID_NAME_HOOK;
#define PID_NAME_HOOK_SIZE (sizeof(PID_NAME_HOOK) - 256)


// itrace     
     
/* the load/store hook could possibly use all 7 parms */
typedef struct STRUCT_ITRACE_INFO_HOOK {
        uint16_t  typeLength;
        uint16_t  majorCode;
        uint32_t  minorCode;
        uint32_t  timeStamp;
        uint32_t  parm1;
        uint32_t  parm2;
        uint32_t  parm3;
        uint32_t  parm4;
        uint32_t  parm5;
        uint32_t  parm6;
        uint32_t  parm7;
} ITRACE_INFO_HOOK;

#define ITRACE_OFF_HOOK_SIZE         20
#define ITRACE_PROC_SWITCH_HOOK_SIZE 20
#define ITRACE_ON_HOOK_SIZE          28
#define ITRACE_BRANCH_HOOK_SIZE      28
#define ITRACE_INST_VSID_HOOK_SIZE   28

#define ITRACE_INFO_HOOK_SIZE 20  	// branch hook, 32-bit
#define ITRACE_INFO_HOOK_SIZE2 28	// branch hook, 64-bit

#define PI_ITRACE_MINOR_STORE       1
#define PI_ITRACE_MINOR_RELATIVE_IP 2
#define PI_ITRACE_MINOR_RELATIVE_LS 4
#define PI_ITRACE_MINOR_VSID        8
#define PI_ITRACE_MINOR_REPEAT_VSID 16
#define PI_ITRACE_MINOR_DATA_LENGTH 32
#define PI_ITRACE_MINOR_INST_VSID   64

#define PI_ITRACE_FLAG_RING_LEVEL   0x00000001
#define PI_ITRACE_FLAG_INST_COUNT   0x00000002
#define PI_ITRACE_FLAG_SS           0x00000004
#define PI_ITRACE_FLAG_VSID         0x00000008
#endif // STAP_ITRACE DEFS

#define PIDS_ITRACE_ENABLE    1
#define PIDS_ITRACE_ACTIVE    2


/* Some defines for Major  codes.  */
#define INTERRUPT_MAJOR         0x04
#define TPROF_MAJOR             0x10
#define PTHREAD_MAJOR           0x30
#define DISPATCH_MAJOR          0x12
#ifndef STAP_ITRACE // DEFS
#define SYS_INFO_MAJOR          0x11
#define TASK_SWITCH_MAJOR       0x13
#define MTE_MAJOR               0x19
#define MODULE_MAJOR            0x20
#define BRANCH_MAJOR            0x20
#define BRANCH_PRIV_MAJOR       0x21
#define PROC_SWITCH_MAJOR       0x2c
#define TRACE_ON_MAJOR          0x2e
#define TRACE_OFF_MAJOR         0x2f
#define PTHREAD_MAJOR           0x30
#define LOAD_STORE_MAJOR        0x31
#define MEMORY_MAJOR            0x40


/* Special Hook Types */
#define   SFT_HOOK_SPECIAL_MAJ  (0xA8)        /* available A8, 1-15 */
#define   SFT_HOOK_TIMEHI_MIN   (0x01)

// Sysinfo minors

  #define SFT_HOOK_SYSINFO_CPUSPEED_MINOR  0xB5
  #define SFT_HOOK_SYSINFO_TIME_BASE_MINOR 0xB7                        
  #define SFT_HOOK_SYSINFO_OS              0xBE
  #define SFT_HOOK_SYSINFO_START_MINOR     0xB4
  #define SFT_HOOK_SYSINFO_END_MINOR       0xB3
#endif // STAP_ITRACE DEFS
  #define SFT_HOOK_SYSINFO_DD_ADDR         0xB6
  #define SFT_HOOK_SYSINFO_SYSTEM_PIDS     0xBA
  #define SFT_HOOK_METRIC_VALUES           0xBC
  #define SFT_HOOK_TPROF_MODE              0xB8
  #define SFT_HOOK_TPROF_EVENT_INFO        0xB9
  #define SFT_HOOK_SYSINFO_PERFDD          0xBD
  #define SFT_HOOK_SYSINFO_CPU             0xBF
  #define SFT_HOOK_TRACE_INTERVAL_ID       0xB2


#define LTPROF_MINOR              0x00000010
#ifndef STAP_ITRACE // DEFS
#define MTE_MINOR		  0x00000001
#define FORK_MINOR		  0x00000063
#define PID_NAME_MINOR            0x00000034
#endif // STAP_ITRACE DEFS
#define IRQ_ENTRY_MINOR           0x00000001
#define IRQ_EXIT_MINOR  	  0x80000000
#define NEWPID_MINOR              0x00000064
#define NOTIFY_EXIT_START_MINOR	  0x00000077
#define NOTIFY_EXIT_END_MINOR	  0x80000077
#define NOTIFY_UNMAP_START_MINOR  0x00000078
#define NOTIFY_UNMAP_END_MINOR	  0x80000078
#define NOTIFY_UNLOAD_START_MINOR 0x00000079
#define NOTIFY_UNLOAD_END_MINOR	  0x80000079
#define NOTIFY_FORK_START_MINOR   0x0000007a
#define NOTIFY_FORK_END_MINOR	  0x8000007a

#define EXIT_HOOK                 0x80000000

/* Some defines related to writing hook records */
/* Raw File Format Hook Types */

#ifndef STAP_ITRACE // DEFS
#define PERF_32BIT           (0x0000)
#define PERF_VARDATA         (0x0001)
#define PERF_MIXED_DATA      (0x0002)
#define PERF_EXTENDED        (0x0003)

#define RAWFMT_HOOK_LEN_MASK (0xfffc)
#endif // STAP_ITRACE DEFS
#define RAWFMT_HOOK_MASK     (0x0003)

#define TPROF_HOOKS      255
#define TS_PERFCLK       255      // time stamp

#ifndef STAP_ITRACE // DEFS
#define COLLECTION_MODE_MTE_END   1
#endif // STAP_ITRACE DEFS
#define COLLECTION_MODE_NORMAL    2
#define COLLECTION_MODE_WRAP      4
#define COLLECTION_MODE_CONTINUOS 8 

/* constants for header fields */
#define	INST_COUNT_ON 	0x00000001


//
// SWTRACE hook sizes
//
#define MAX_NRM2_HOOK_SIZE        (0xFFFC)   // 65532
#define FAST_NRM2_HOOK_SIZE       (0xFA0)    // 4000

#define NRM2_HOOK_HEADER_SIZE           12   // NRM2: hook header (includes TS)
#define NRM2_TSCHANGE_HOOK_DATASIZE      4   // NRM2: TSC upper 32 bits data bytes
#define NRM2_DISPATCH_HOOK_DATASIZE      4   // NRM2: 4Byte TID (4Byte PID is minor code)
#if defined(_X86)
#define NRM2_INTERRUPT_HOOK_DATASIZE     4   // NRM2: 4Byte IntrFrame
#define NRM2_TPROF_HOOK_DATASIZE        12   // NRM2: 4Byte PID, 4Byte TID, 4Byte addr (SegAttr ORed in to minor code)
#else
#define NRM2_INTERRUPT_HOOK_DATASIZE     0   // NRM2: (IRQ is minor code)
#define NRM2_TPROF_HOOK_DATASIZE        16   // NRM2: 4Byte PID, 4Byte TID, 8Byte addr (SegAttr ORed in to minor code)
#endif

//
// SWTRACE major/minor codes
//
#define SFT_HOOK_EXCEPTION_MAJOR        0x03

#define SFT_HOOK_INTR_MAJOR             0x04

#define SFT_HOOK_SYSSERVICE_MAJOR       0x05
#define SFT_HOOK_SYSCALL_MAJOR          SFT_HOOK_SYSSERVICE_MAJOR

#define SFT_HOOK_TSCHANGE_MAJOR         0x08
  #define SFT_HOOK_CYCLES_CHANGE_MAJOR    0x01

#define SFT_HOOK_METRICS_MAJOR          0x09

#define SFT_HOOK_TPROF_MAJOR            0x10
  #define SFT_HOOK_TPROF_MINOR            0x20  // Needs to be ORed with the CPL


#define SFT_HOOK_DISP_MAJOR            0x12

#define SFT_HOOK_MTE_MAJOR             0x19
  #define SFT_HOOK_MTE_MINOR              0x01
  #define SFT_HOOK_PIDNAME_MINOR          0x34
  #define SFT_HOOK_3264BIT_MINOR          0x35
  #define SFT_HOOK_CLONE_FORK_MINOR       0x63
  #define SFT_HOOK_CREATE_PID_MINOR       0x64
  #define SFT_HOOK_KILL_PID_MINOR         0x65
  #define SFT_HOOK_CREATE_THR_MINOR       0x66
  #define SFT_HOOK_KILL_THR_MINOR         0x67

#define SFT_HOOK_ITRACE_BR_FROM_RING0  0x20
#define SFT_HOOK_ITRACE_BR_FROM_RING3  0x21
#define SFT_HOOK_ITRACE_RING0_INTR     0x28
#define SFT_HOOK_ITRACE_RING3_INTR     0x29
#define SFT_HOOK_ITRACE_PIDTID_CHANGE  0x2C
#define SFT_HOOK_ITRACE_TRACE_START    0x2E
#define SFT_HOOK_ITRACE_TRACE_END      0x2F

#define SFT_HOOK_HOOKIT_MAJOR          0xB0
  #define SFT_HOOK_HOOKIT_ENTRY_MINOR     0x00
  #define SFT_HOOK_HOOKIT_EXIT_MINOR      0x80

//
// NRM2 hook types
//
//#define RAWFMT_HOOK_LEN_MASK    (0xFFFC)
//#define RAWFMT_HOOK_MASK        (0x0003)
#define RAWFMT_HOOK_32BIT       (0x0000)    // n 32-bit values
#define RAWFMT_HOOK_32BIT_NOTS  (0x0001)    // n 32-bit values, no timestamp
#define RAWFMT_HOOK_VARDATA     (0x0001)    // n buffers ***** NO LONGER USED *****
#define RAWFMT_HOOK_MIXED_DATA  (0x0002)    // n 32-bit values and m buffers
#define RAWFMT_HOOK_EXTENDED    (0x0003)    // user hook

//
// RAW NRM2 Hook
//
// HookData varies by hook type ...
//
// 32BIT:
//          ULONG num_ints;     // Number of ints to follow
//          ULONG ints[*];      // Array of ints
//
// VARDATA:
//          VDATA bufs[*];      // Array of (int, char *) pairs
//
// MIXED:
//          ULONG num_ints;     // Number of ints to follow
//          ULONG ints[*];      // Array of ints
//          VDATA bufs[*];      // Array of (int, char *) pairs
//
struct _RAWFMT_HOOK {
   USHORT  TypeLen;                         // Bits 15-2 are length, bits 1-0 are type
   USHORT  Major;                           // Major code
   UINT32  Minor;                           // Minor code
   UINT32  TimeStamp;                       // Timestamp (aka. metric #1)
   UCHAR   HookData[1];                     // Varies with hook type and number of metrics
};
typedef struct _RAWFMT_HOOK  RAWFMT_HOOK;

#endif
