/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2008
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef POST_H
   #define POST_H

   #include "encore.h"

   #include <stdio.h>
   #include <stdarg.h>

// atoll() reads int64 value on Linux and AIX, _atoi64() on Windows
// fseeko() moves the file pointer on 64-bit position on Linux,
// _fseeki64() on Windows

   #if defined(_WINDOWS)
      #define atoll  _atoi64
//     #define fseeko _fseeki64  // requires Visual Studio 2005
//     #define ftello _ftelli64  // requires Visual Studio 2005
      #define ftello ftell
   #endif    

//==== Variable Message Handling routines ==============================

   #define MVMsg(x,y) if ( 0 != ( gv.mask & x ) ) OptVMsgLog y

size_t OptVMsg(                const char *fmt, ... );
size_t OptVMsgLog(             const char *fmt, ... );
size_t ErrVMsgLog(             const char *fmt, ... );
size_t WriteFlush( FILE *file, const char *fmt, ... );

size_t MaskedVMsg( long  mask, const char *fmt, ... );   // Deprecated

//==== Variable Message Handling routines ==============================

   #define MAX_BUFFER_SIZE  1024

///////////////////////////
//// tprof TNode ////
typedef struct _TNode   TNode;
typedef struct _TNode * pTNode;

//   type      char   id     name
//   ========  =====  =====  ======
//   pid       p      pid    sprintf("0x%4x_%s", pid, pnm)
//   tid       t      tid    sprintf("0x%4x_%s", tid, tnm)
//   mod       m      0      mod
//   sym       n      0      sym
//   offset    ?      ?      ?

// Node needs to be redesigned for space
// node content => void *
// only fillin for symbol Node
struct _TNode
{
   UINT64 sttaddr;                      // non-zero *only* for symbol nodes

   int    base;                         // ticks
   int    cum;                          // cum ticks

   char * nm;                           // ascii string (pnm, snm, sym, ...)

   // NB: void * data candidates( inlined for max memory usage )
   // need DEEP copies
   char * cptr;                         // code ptr (to start of symbol)
   int    clen;                         // code len

   // ENIO Warning.
   //  overloading rc : symNode     => quality
   //                 :  offsetNode => line_no
   int    rc;
   char * snm;                          // sym src_file ptr
   void * module;                       // a2n shortcuts for late data
   void * symbol;                       // a2n shortcuts for late data
   // end void * stuff

   unsigned int  flags;                 // Flags from SYMDATA

   int    ccnt;                         // no of children ????
   pTNode par;                          // parent
   pTNode chd;                          // 1st child
   pTNode sib;                          // next sibling
};

pTNode push(pTNode p, char * nm);
pTNode pusha(pTNode p, char * nm, uint64 addr);
pTNode initTree(void);
pTNode dualx;
int dlev[5];

///////////////////////////
//// post rt-tree Node ////
typedef struct _Node   Node;
typedef struct _Node * pNode;

struct _Node
{
   int  base;
   int  calls;

   char * mnm;                          // method
   char * mod;                          // module

   int    ccnt;                         // no of children
   pNode  par;                          // parent
   pNode  chd;                          // 1st child
   pNode  sib;                          // next sibling
};

typedef struct _gd GD;
struct _gd
{
   FILE * tp;
   FILE * tpaddr;
   FILE * arc;
   FILE * xfd;
   FILE * msg;
// FILE * ptree;
   FILE * rw;                           // rewrite wo ts (compression)
   FILE * show;                         // show nrm in ascii
   FILE * micro;                        // show nrm in ascii
   FILE * qtrace;                       // qtrace output
   FILE * wi;                           // qtrace output
   FILE * ls;                           // load/store trace
};
extern GD gd;

typedef struct _gv GV;
struct _gv
{
   double   clip;                       // min tks % to show(def 1)
   int64    tstt;
   int64    tstp;

   double   clock_rate;                 // Cycles per second
   double   catz;                       // Cycles At Time Zero (AIX Only)

// Debug output mask definitions

   #define gv_mask_Show_Disasm         0x00000100
   #define gv_mask_Show_Files          0x00000200

   long     mask;                       // debug output mask

   #define gv_verbose_stderr  1         // Optional messages to stderr
   #define gv_verbose_logmsg  2         // Optional messages to logmsg, if possible
   #define gv_verbose_maximum 3         // Optional messages to stderr and logmsg

   int      verbose;                    // Level of verbosity, 0 = no optional messages

//-------------------------------

   int      aixtrc;                     // aix trace
   int      trcid;                      // Sequence number for JITA2N
   int      oldjit;                     // debug to force old jit algorithm

   // common
   int      readbytes;
   char   * sympath;                    // alternate search path for symbols
   int      undup;                      // separate sym by adding stt_addr suffix
   int      oprof;                      // oprofile (i.e., Linux), two trace passes (1 to get mte)
   int      trace_type;                 // Normal, continuous, or wrap-around (Linux only)

   // tprof
   int      tpsel;                      // sel 1 of 255 trof hooks
   int      itsel;                      // sel 1 of 255 metric fields from a itrace hook
   int      twidth;                     // tprof tk field width
   int      micro;                      // micro profiling output
   int      tprof;                      // any tprof hooks => 1
   int      tottks;
   int      ntonly;
   int      nospace;
   char   * flab[8];                    // func labels in compressed tree
   char   * rlab[8];                    // title labels
   int      llev;                       // last tree out level
   int      rsel;                       // report selection vector
   char   * rselvec;                    // report selcection vector
   int      xlevs;                      // levels in xtra tree
   int      xind[5];                    //
   char     xlabs[5][4];                // x lev labs
   pTNode   stk[5];                     // stack of data
   FILE   * xfd;                        // eXtra reports
   int      pnmonly;
   int      tnmonly;
   int      cpu;                        // include cpu into pid for reports
   int      off;                        // offset report

   int      tprof_mode;                 // Tprof mode (0: time-based, 1: event-based )
   int      tprof_compat;               // Compatibility mode - do not print the mode line
   int      tprof_event;                // Event number for event-based
   int      tprof_count;                // Tprof period in microseconds for time-based,
                                        // event count for event-based

   // call flow
// int      ptree;                      // callflow ptree file (ala log-rt)
   int      cstats;                     // compaction stats
   int      arc;                        // callflow arc file   (ala log-gen)
   int      cfcnt;                      // callflow hooks used
   int      aatot;                      // total callflow hooks
   int      ring0;                      // 0: none 1: sum 2: all
   int      mtrace;                     // some ss/cf hooks
   int      showss;                     // ss:1 other:0
   int      compact;                    // compact mode
   int      v;                          // verbose flag
   int      nkbar;                      // extra label in/out of ring3(kernel)
   int      stap;                       // convert stap data file to nrm2   

   int      db;
   int      nv;                         // dont validate symbols
   int      c2pid[64];
   int      maxcpu;
   int      nmcnt;                      // NoMod Count
   int      nmstp;                      // stop after nmstp NoMod events
   int      instrs;                     // instrs in ss format
   int      showx;                      // Only show nrm2, no processing
   int      noskew;                     // no clock skew compensation
   int      a2nopt;                     // reuse a2n calls within same Sym(rc = 0) only
   int      timebase;                   // ppc. trace time base different from MHz
   int      a2ndbg;                     // full a2n debug
   int      tmod;                       // tag Module with _32bit or _64bit
   int      a2nrdup;                    // rename duplicate symbols
   int      disasm;                     // Disassembly in tprof.micro, 0 = ascii only
};
extern GV gv;

// remember jita2n data from files
typedef struct _jrec JREC;

   #define jrec_cflag_Code    0x0001
   #define jrec_cflag_MMI     0x0002

struct _jrec
{
   JREC  * next;                        // Pointer to next JREC in list
   int     size;                        // Size of code
   uint64  addr;                        // Address of code when executing
   int     tlo;                         // Time stamp (low)
   int     thi;                         // Time stamp (high)
   char  * mnm;                         // Method name
   char  * code;                        // Address of saved bytecodes
   int     cflag;                       // Flags: uses jrec_flags_????
   int     seqNo;                       // Sequence number from driver
};

// remember jtnm data from files
typedef struct _jtnm JTNM;

struct _jtnm
{
   JTNM   * next;                       // Pointer to next record in list
   char   * tnm;                        // Thread name ("" if END event)
   unsigned int tlo;                    // Time stamp (low) 
   unsigned int thi;                    // Time stamp (high)
   int    tid;                          // Thread ID
   int    tnmid;                        // 0 or trcid when written
};

char ** tokenize(char * str, int * pcnt);
void    tree_sort(pTNode cx, int lev);
pTNode  tree_sort_out_free(pTNode cx, int lev, FILE * fd);
void    tree_new(pTNode n, int lev, pTNode root, int tklev);
void    tree_cum(pTNode cx, int lev);
void    xtreeOut(char * fnm);
INT64   nrm2_size(void);
INT64   getReadBytes(void);
int     noStopOnInvalid(void);
UINT64  getTraceTime(void);

FILE  * wopen( char * fnm, char * mode );
FILE  * xopen( char * fnm, char * mode );

extern char * i386_instruction( char *data, uint64 addr, int *, char *, int disasm_mode);

#endif
