/*   IBM Performance Inspector
 *   Copyright (c) International Business Machines Corp., 2003 - 2007
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <qtlib.h>

unsigned short flags_for_rec(qtrace_rec *);

int qtlib_header(uint64_t addr, FILE *outfile)
{
	uint64_t ul;
	unsigned int ui;
	unsigned short us;

	/* beginning */
	ul = 0x00010002;
	fwrite(&ul, sizeof(uint64_t), 1, outfile);

	/* flags */
	us = 0xe000;
	fwrite(&us, sizeof(unsigned short), 1, outfile);

	/* magic */
	ui = 0;
	fwrite(&ui, sizeof(unsigned int), 1, outfile);

	/* version */
	ui = 0x01040700;
	fwrite(&ui, sizeof(unsigned int), 1, outfile);

	/* iar */
    ul = addr;
    fwrite(&ul, sizeof(uint64_t), 1, outfile);

	return 0;
}

int qtlib_record(qtrace_rec *rec, FILE *outfile)
{
	uint64_t ul;
	unsigned int ui;
	unsigned short us;
	unsigned char ub;

	/* instruction */
	ui = rec->instruction;
	fwrite(&ui, sizeof(unsigned int), 1, outfile);

	/* flags? cwc */
	us = flags_for_rec(rec);
	fwrite(&us, sizeof(unsigned short), 1, outfile);

	if (rec->flags & QTRACE_REC_BRANCH) {
		ub = 0; /* node */
		fwrite(&ub, sizeof(unsigned char), 1, outfile);
		us = QTRACE_ExceededMaximumBranchDepth; /* termination_code */
		fwrite(&us, sizeof(unsigned short), 1, outfile);
	}
	if (rec->flags & QTRACE_REC_LOAD_STORE) {
		ul = rec->data_ea;
		fwrite(&ul, sizeof(uint64_t), 1, outfile);
	}
	if (rec->flags & QTRACE_REC_DATA_VSID) {
		ul = *((uint64_t *) &rec->data_vsid);
		fwrite(&ul, 7, 1, outfile); /* cwc len? */
	}
	if (rec->flags & QTRACE_REC_LENGTH) {
		ub = rec->length;
		fwrite(&ub, sizeof(unsigned char), 1, outfile);
	}
	if ((rec->flags & QTRACE_REC_BRANCH) &&
		!(rec->flags & QTRACE_REC_UNTAKEN_BRANCH)) {
		ul = rec->instruction_ea;
		fwrite(&ul, sizeof(uint64_t), 1, outfile);
	}
	return 0;
}

unsigned short flags_for_rec(qtrace_rec *rec)
{
	unsigned short rv = 0;

	if (rec->flags & QTRACE_REC_BRANCH) {
		rv |= 0x6000; /* node_present(1) termination_present(2) */

		if (!(rec->flags & QTRACE_REC_UNTAKEN_BRANCH)) {
			/* iar_change_present(0) instruction_address_present(9) */
			rv |= 0x8040; 
		}
	}

	if (rec->flags & QTRACE_REC_LOAD_STORE) {
		rv |= 0x0800; /* data_address_present(4) */
		if (rec->flags & QTRACE_REC_DATA_VSID)
			rv |= 0x0400; /* data_vsid_present(5) */
		if (rec->flags & QTRACE_REC_LENGTH) {
			rv |= 0x0100; /* length_present(7) */
		}
	}
	return rv;
}
